import hljs from 'highlight.js';

/**
 * Converts raw Markdown into tokens
 */
function tokenize(input) {
    const lines = input.split("\n");
    const tokens = [];
    let i = 0;
    while (i < lines.length) {
        const line = lines[i];
        const trimmedLine = line.trim();
        if (!trimmedLine) {
            i++;
            continue; // Skip empty lines
        }
        const token = detectToken(trimmedLine, lines, i);
        if (token) {
            tokens.push(token.token);
            i = token.newIndex;
        }
        else {
            tokens.push({ type: "paragraph", content: trimmedLine });
            i++;
        }
    }
    return tokens;
}
function detectToken(line, lines, index) {
    const headingToken = detectHeading(line);
    if (headingToken)
        return { token: headingToken, newIndex: index + 1 };
    const hrToken = detectHorizontalRule(line);
    if (hrToken)
        return { token: hrToken, newIndex: index + 1 };
    const listToken = detectList(lines, index);
    if (listToken)
        return listToken;
    const blockquoteToken = detectBlockquote(lines, index);
    if (blockquoteToken)
        return blockquoteToken;
    const imageToken = detectImage(line);
    if (imageToken)
        return { token: imageToken, newIndex: index + 1 };
    const linkToken = detectLink(line);
    if (linkToken)
        return { token: linkToken, newIndex: index + 1 };
    const tableToken = detectTable(lines, index);
    if (tableToken)
        return tableToken;
    const codeBlockToken = detectCodeBlock(lines, index);
    if (codeBlockToken)
        return codeBlockToken;
    return null;
}
function detectList(lines, index) {
    const orderedMatch = /^\d+\.\s+(.*)$/.exec(lines[index].trim());
    const unorderedMatch = /^[-*]\s+(.*)$/.exec(lines[index].trim());
    if (orderedMatch || unorderedMatch) {
        const items = [];
        const isOrdered = !!orderedMatch;
        while (index < lines.length && (isOrdered ? /^\d+\.\s+(.*)$/.test(lines[index].trim()) : /^[-*]\s+(.*)$/.test(lines[index].trim()))) {
            items.push(lines[index].trim().replace(isOrdered ? /^\d+\.\s+/ : /^[-*]\s+/, ""));
            index++;
        }
        return { token: { type: isOrdered ? "ordered_list" : "unordered_list", items }, newIndex: index };
    }
    return null;
}
function detectHeading(line) {
    const match = /^(#{1,6})\s+(.*)$/.exec(line);
    if (match) {
        return {
            type: "heading",
            content: match[2],
            level: match[1].length,
        };
    }
    return null;
}
function detectBlockquote(lines, index) {
    if (lines[index].trim().startsWith(">")) {
        const content = [];
        while (index < lines.length && lines[index].trim().startsWith(">")) {
            content.push(lines[index].trim().replace(/^>\s*/, ""));
            index++;
        }
        return {
            token: { type: "blockquote", content: content.join("\n") },
            newIndex: index,
        };
    }
    return null;
}
function detectImage(line) {
    const match = /^!\[([^\]]*)\]\(([^)]+)\)\.?$/.exec(line);
    if (match) {
        return {
            type: "image",
            alt: match[1],
            src: match[2],
        };
    }
    return null;
}
function detectLink(line) {
    const match = /^\[([^\]]+)\]\(([^)]+)\)\.?$/.exec(line);
    if (match) {
        return {
            type: "link",
            text: match[1],
            href: match[2],
        };
    }
    return null;
}
function detectTable(lines, index) {
    const tableRows = [];
    while (index < lines.length) {
        const line = lines[index].trim();
        // Stop if we encounter a non-empty line that doesn't resemble a table row
        if (!line.startsWith("|") || !line.endsWith("|")) {
            // Allow incomplete rows (e.g., "| Missing Header")
            if (line.startsWith("|")) {
                tableRows.push(line);
            }
            else {
                break;
            }
        }
        else {
            tableRows.push(line);
        }
        index++;
    }
    if (tableRows.length > 0) {
        return {
            token: { type: "table", content: tableRows.join("\n") },
            newIndex: index,
        };
    }
    return null;
}
function detectCodeBlock(lines, index) {
    if (lines[index].trim().startsWith("```")) {
        const codeLines = [];
        const language = lines[index].trim().slice(3).trim(); // Capture the optional language
        let i = index + 1;
        while (i < lines.length && !lines[i].trim().startsWith("```")) {
            codeLines.push(lines[i]);
            i++;
        }
        if (i < lines.length && lines[i].trim().startsWith("```")) {
            // Successfully found the closing ```
            return {
                token: {
                    type: "code_block",
                    content: codeLines.join("\n"),
                    language,
                },
                newIndex: i + 1,
            };
        }
    }
    return null;
}
function detectHorizontalRule(line) {
    // Matches horizontal rules with 3+ hyphens, asterisks, or underscores
    const hrRegex = /^ {0,3}([-*_])(?: *\1){2,} *$/;
    return hrRegex.test(line) ? { type: "horizontal_rule" } : null;
}

/**
 * Parses tokens into HTML
 */
function parse(tokens, plugins = []) {
    return tokens
        .map((token) => {
        // Allow plugins to process the token
        for (const plugin of plugins) {
            const result = plugin(token);
            if (result)
                return result;
        }
        // Default rendering logic
        switch (token.type) {
            case "heading":
                return renderHeading(token);
            case "paragraph":
                return renderParagraph(token);
            case "unordered_list":
                return renderUnorderedList(token);
            case "ordered_list":
                return renderOrderedList(token);
            case "blockquote":
                return renderBlockquote(token);
            case "table":
                return renderTable(token);
            case "image":
                return renderImage(token);
            case "link":
                return renderLink(token);
            case "code_block":
                return renderCodeBlock(token);
            case "horizontal_rule":
                return renderHorizontalRule();
            default:
                return ""; // Fallback for unhandled token types
        }
    })
        .join("");
}
function renderHeading(token) {
    if (token.content && token.level) {
        return `<h${token.level}>${token.content}</h${token.level}>`;
    }
    return "";
}
function renderParagraph(token) {
    if (token.content) {
        const content = token.content.replace(/!\[([^\]]*)\]\(([^)]+)\)|\[([^\]]+)\]\(([^)]+)\)/g, (_, alt, src, text, href) => alt && src
            ? `<img alt="${alt}" src="${src}" />`
            : text && href
                ? `<a href="${href}">${text}</a>`
                : "");
        return `<p>${content}</p>`;
    }
    return "";
}
function renderUnorderedList(token) {
    if (token.items && Array.isArray(token.items)) {
        const listItems = token.items.map((item) => `<li>${item}</li>`).join("");
        return `<ul>${listItems}</ul>`;
    }
    return "";
}
function renderOrderedList(token) {
    if (token.items && Array.isArray(token.items)) {
        const listItems = token.items.map((item) => `<li>${item}</li>`).join("");
        return `<ol>${listItems}</ol>`;
    }
    return "";
}
function renderBlockquote(token) {
    if (token.content) {
        const content = token.content
            .split("\n") // Split blockquote into individual lines
            .map((line) => `<p>${line}</p>`)
            .join(""); // Wrap each line in a paragraph
        return `<blockquote>${content}</blockquote>`;
    }
    return "";
}
function renderTable(token) {
    if (token.content) {
        const rows = token.content.trim().split("\n");
        const headers = rows[0]
            .split("|")
            .filter((header) => header.trim() !== "")
            .map((header) => `<th>${header.trim()}</th>`)
            .join("");
        const body = rows
            .slice(1) // Skip the header
            .map((row) => `<tr>${row
            .split("|")
            .filter((cell) => cell.trim() !== "")
            .map((cell) => `<td>${cell.trim()}</td>`)
            .join("")}</tr>`)
            .join("");
        return `<table><thead><tr>${headers}</tr></thead><tbody>${body}</tbody></table>`;
    }
    return "";
}
function renderImage(token) {
    if (token.src) {
        const altText = token.alt || ""; // Use empty alt text if undefined
        return `<img src="${token.src}" alt="${altText}" />`;
    }
    return "";
}
function renderLink(token) {
    if (token.href && token.text) {
        return `<a href="${token.href}">${token.text}</a>`;
    }
    return "";
}
function renderCodeBlock(token) {
    if (token.content) {
        return `<pre><code>${escapeHtml$1(token.content)}</code></pre>`;
    }
    return "";
}
// Add this new renderer function
function renderHorizontalRule() {
    return "<hr>\n";
}
function escapeHtml$1(content) {
    return content
        .replace(/&/g, "&amp;")
        .replace(/</g, "&lt;")
        .replace(/>/g, "&gt;")
        .replace(/\"/g, "&quot;")
        .replace(/'/g, "&#39;");
}

function escapeHtml(str) {
    const entityMap = {
        "<": "&lt;",
        ">": "&gt;",
        "&": "&amp;",
        '"': "&quot;",
        "'": "&#39;",
    };
    return str.replace(/[<>&"']/g, (match) => entityMap[match]);
}

function syntaxHighlightPlugin(token) {
    console.log(token);
    if (token.content) {
        const language = token.language || "plaintext"; // Default to plaintext if no language is specified
        // Check if the language is supported by highlight.js
        if (hljs.getLanguage(language)) {
            // Highlight the code using highlight.js
            const highlightedCode = hljs.highlight(token.content, { language }).value;
            // Return the highlighted code as HTML
            return `<pre class="hljs"><code class="language-${language}">${highlightedCode}</code></pre>`;
        }
        else {
            // Fallback for unsupported languages or plaintext
            return `<pre class="hljs"><code>${escapeHtml(token.content)}</code></pre>`;
        }
    }
    return null; // Return null if there's no content
}
/**
 * Table plugin to convert table tokens to HTML
 */
function tablePlugin(token) {
    if (token.type === "table" && token.content) {
        const rows = token.content
            .trim()
            .split("\n")
            .map((row) => row.trim())
            .filter((row) => row.length > 0);
        if (rows.length < 2) {
            return null;
        }
        const headers = rows[0]
            .split("|")
            .map((header) => header.trim())
            .filter((header) => header !== "")
            .map((header) => `<th>${header}</th>`)
            .join("");
        const body = rows
            .slice(2)
            .map((row) => `<tr>${row
            .split("|")
            .map((cell) => cell.trim())
            .filter((cell) => cell !== "")
            .map((cell) => `<td>${cell}</td>`)
            .join("")}</tr>`)
            .join("");
        return `<table><thead><tr>${headers}</tr></thead><tbody>${body}</tbody></table>`;
    }
    return null;
}
function inlineStylesPlugin(token) {
    if (token.type === "paragraph" &&
        token.content &&
        (/<.*?>/.test(token.content) ||
            /\*\*.*?\*\*|__.*?__|\*.*?\*|_.*?_|~~.*?~~|`.*?`/.test(token.content))) {
        let content = token.content;
        content = escapeHtml(content);
        content = content.replace(/(\*\*|__)(.*?)\1/g, "<strong>$2</strong>");
        content = content.replace(/(\*|_)(.*?)\1/g, "<em>$2</em>");
        content = content.replace(/~~(.*?)~~/g, "<del>$1</del>");
        content = content.replace(/`(.*?)`/g, "<code>$1</code>");
        return `<p>${content}</p>`;
    }
    return null;
}
function imagePlugin(token) {
    if (token.type === "paragraph" &&
        token.content &&
        /!\[([^\]]*)\]\(([^)]+)\)/.test(token.content)) {
        const res = token.content.replace(/!\[([^\]]*)\]\(([^)]+)\)/g, (_, altText, src) => `<img src="${src}" alt="${altText || ""}">`);
        return `<p>${res}</p>`;
    }
    return null;
}
function linkPlugin(token) {
    if (token.type === "paragraph" &&
        token.content &&
        /\[([^\]]+)\]\(([^)]+)\)/.test(token.content)) {
        const res = token.content.replace(/\[([^\]]+)\]\(([^)]+)\)/g, (_, text, href) => `<a href="${href}" target="_blank" rel="noopener noreferrer">${text}</a>`);
        return `<p>${res}</p>`;
    }
    return null;
}

/**
 * Markdown parser class
 */
class MarkdownParser {
    plugins;
    useSyntaxHighlighting;
    inCodeBlock;
    buffer = "";
    currentList; // Add this line to define the currentList property
    constructor() {
        // Initialize with default plugins (excluding syntaxHighlightPlugin)
        this.plugins = [
            tablePlugin.bind(this),
            inlineStylesPlugin.bind(this),
            imagePlugin.bind(this),
            linkPlugin.bind(this),
        ];
        // Syntax highlighting is disabled by default
        this.useSyntaxHighlighting = false;
        this.inCodeBlock = false;
        this.currentList = null;
        this.reset();
    }
    /**
     * Allows additional plugins to be registered
     * @param plugin A function that processes a token and returns HTML or null
     */
    use(plugin) {
        this.plugins.push(plugin);
        return this;
    }
    /**
     * Enables or disables syntax highlighting
     * @param enable A boolean indicating whether to enable syntax highlighting
     */
    enableSyntaxHighlighting(enable) {
        this.useSyntaxHighlighting = enable;
        if (enable) {
            // Add syntaxHighlightPlugin if not already present
            if (!this.plugins.includes(syntaxHighlightPlugin)) {
                this.plugins.push(syntaxHighlightPlugin.bind(this));
            }
        }
        else {
            // Remove syntaxHighlightPlugin if it exists
            this.plugins = this.plugins.filter((plugin) => plugin !== syntaxHighlightPlugin);
        }
        return this;
    }
    /**
     * Renders Markdown input into HTML
     * @param input The Markdown string to be rendered
     * @returns Rendered HTML string
     */
    render(input) {
        // Tokenize the input
        const tokens = tokenize(input);
        // Parse the tokens into HTML
        return parse(tokens, this.plugins);
    }
    /**
     * Resets the parser to its initial state
     */
    // New methods for streaming support
    reset() {
        // Initialize all stateful properties
        this.inCodeBlock = false;
        this.currentList = null;
        this.buffer = "";
        // Add other stateful properties from your original implementation
    }
    /**
     * Parses a chunk of Markdown text
     * @param chunk A chunk of Markdown text
     * @returns The parsed HTML string
     */
    // New streaming-friendly parse method
    parseChunk(chunk) {
        this.buffer += chunk;
        const lines = this.buffer.split("\n");
        this.buffer = lines.pop() || "";
        let html = "";
        for (const line of lines) {
            html += this._parseLine(line);
        }
        return html;
    }
    finalize() {
        let html = "";
        if (this.buffer) {
            html += this._parseLine(this.buffer);
        }
        // Close any remaining open blocks
        if (this.inCodeBlock)
            html += "</code></pre>";
        this.reset();
        return html;
    }
    tokenizeStream(line) {
        // Implement your actual tokenization logic here
        const tokens = [];
        // Example simple tokenization:
        if (line.startsWith("```")) {
            tokens.push({
                type: "code_block",
                content: line.trim(),
                language: line.replace(/```/g, "").trim(),
            });
        }
        else if (line.startsWith("#")) {
            const level = line.match(/#/g)?.length || 1;
            tokens.push({
                type: "heading",
                content: line.replace(/#/g, "").trim(),
                level: Math.min(level, 6),
            });
        }
        else {
            tokens.push({ type: "paragraph", content: line });
        }
        return tokens;
    }
    // Private method for line parsing
    _parseLine(line) {
        const tokens = this.tokenizeStream(line);
        // console.log("token-->", tokens);
        // Your existing parsing logic adapted for single lines
        // ... modified to use this.inCodeBlock, this.currentList etc ...
        let lineHtml = ""; // Assuming lineHtml is a string
        if (this.inCodeBlock) {
            lineHtml += this.parseCodeBlockLine(tokens);
        }
        else {
            lineHtml += this.parseRegularLine(tokens);
        }
        return lineHtml;
    }
    parseRegularLine(tokens) {
        let html = "";
        for (const token of tokens) {
            switch (token.type) {
                case "heading":
                    html += `<h${token.level}>${token.content}</h${token.level}>`;
                    break;
                case "code_block":
                    this.inCodeBlock = true;
                    html += `<pre><code class="language-${token.language}">`;
                    break;
                case "paragraph":
                    html += this.parseInlineElements(token.content || "");
                    break;
            }
        }
        return html;
    }
    parseInlineElements(text) {
        // Simple inline parsing example
        return `<p>${text
            .replace(/\*\*(.*?)\*\*/g, "<strong>$1</strong>")
            .replace(/\*(.*?)\*/g, "<em>$1</em>")}</p>`;
    }
    parseCodeBlockLine(tokens) {
        let html = "";
        for (const token of tokens) {
            if (token.type === "code_block") {
                this.inCodeBlock = false;
                return "</code></pre>";
            }
            html += `${token.content}\n`;
        }
        return html;
    }
}

class MarkdownStreamer {
    parser;
    constructor() {
        this.parser = new MarkdownParser();
    }
    get transformStream() {
        return new TransformStream({
            transform: (chunk, controller) => {
                const html = this.parser.parseChunk(typeof chunk === "string" ? chunk : new TextDecoder().decode(chunk));
                if (html)
                    controller.enqueue(html);
            },
            flush: (controller) => {
                const finalHtml = this.parser.finalize();
                if (finalHtml)
                    controller.enqueue(finalHtml);
            },
        });
    }
    // Alternative node.js stream interface
    nodeStream() {
        return new (require("stream").Transform)({
            transform: (chunk, encoding, callback) => {
                try {
                    // console.log("nodeStream chunk", chunk.toString());
                    const html = this.parser.parseChunk(chunk.toString());
                    callback(null, html);
                }
                catch (err) {
                    callback(err);
                }
            },
            flush: (callback) => {
                try {
                    const finalHtml = this.parser.finalize();
                    callback(null, finalHtml);
                }
                catch (err) {
                    callback(err);
                }
            },
        });
    }
}

export { MarkdownParser, MarkdownStreamer };
